

//  Conversions from concentrations to forcings and backwards.


//  New calculations following Etminan, M., G. Myhre, E. J. Highwood,
//  and K. P. Shine (2016), Radiative forcing of carbon dioxide,
//  methane, and nitrous oxide: A significant revision of the methane
//  radiative forcing, Geophys. Res. Lett., 43, 12,614–12,623,
//  doi:10.1002/2016GL071930
//  This approximation is valid in the following ranges:
//  CO2: 180-2000 ppm
//  N2O: 200-525 ppb
//  CH4: 340-3500 ppb


// convert CO2 concentration in ppm to radiative forcing...:
func co2_c2f( c, n, c0=, n0= )
{
  
  // default values:
  if( is_void(n) )
    n = 270.;            // preindustrial N2O concentration (ppb)
  if( is_void(c0) )
    c0 = 278.;           // preindustrial CO2 concentration (ppm)
  if( is_void(n0) )
    n0 = 270.;           // preindustrial N2O concentration (ppb)

  if( c<180. | c>2000. | n<200. | n>525. )
    exit, "[co2_c2f] Greenhouse-gas concentrations outside valid range!";

  // constants:
  a1 = -2.4e-7;          // (W/m2/ppm2)
  b1 =  7.2e-4;          // (W/m2/ppm)
  c1 = -2.1e-4;          // (W/m2/ppb)

  // calculate and return radiative forcing:
  n_avg = 0.5*(n+n0);
  return ( a1*(c-c0)^2 + b1*abs(c-c0) + c1*n_avg + 5.36 )*log(c/c0);

}


// the inverse of this function:
func co2_f2c( forc, n, c0=, n0= )
{
  // default values:
  if( is_void(n) )
    n = 270.;            // preindustrial N2O concentration (ppb)
  if( is_void(c0) )
    c0 = 278.;           // preindustrial CO2 concentration (ppm)
  if( is_void(n0) )
    n0 = 270.;           // preindustrial N2O concentration (ppb)

  // compute average N concentration:
  n_avg = 0.5*(n+n0);

  // constants:
  a1 = -2.4e-7;          // (W/m2/ppm2)
  b1 =  7.2e-4;          // (W/m2/ppm)
  c1 = -2.1e-4;          // (W/m2/ppb)

  c_l = 180.;
  c_h = 2000.;
  dc = c_h-c_l;
  
  while( dc>1.e-9 ){
    c = .5*(c_l+c_h);
    val = forc - ( a1*(c-c0)^2 + b1*abs(c-c0) + c1*n_avg + 5.36 )*log(c/c0);
    if( sign(val)>0 )
      c_l = c;
    if( sign(val)<0 )
      c_h = c;
    dc = c_h-c_l;
  }

  return c;
  
}

// convert N2O concentration in ppb to radiative forcing...:
func n2o_c2f( n, c, m, n0=, c0=, m0= )
{
  
  // default values:
  if( is_void(c) )
    c = 278.;            // preindustrial CO2 concentration (ppm)
  if( is_void(m) )
    m = 700.;            // preindustrial CH4 concentration (ppb)
  if( is_void(n0) )
    n0 = 270.;           // preindustrial N2O concentration (ppb)
  if( is_void(c0) )
    c0= 278.;            // preindustrial CO2 concentration (ppm)
  if( is_void(m0) )
    m0 = 700.;           // preindustrial CH4 concentration (ppb)

  if( c<180. | c>2000. | n<200. | n>525. | m<340. | m>3500. )
    exit, "[n2o_c2f] Greenhouse-gas concentrations outside valid range!";

  
  // constants:
  a2 = -8.0e-6;          // (W/m2/ppm)
  b2 =  4.2e-6;          // (W/m2/ppb)
  c2 = -4.9e-6;          // (W/m2/ppb)

  // calculate and return radiative forcing:
  c_avg = 0.5*(c+c0);
  m_avg = 0.5*(m+m0);
  n_avg = 0.5*(n+n0);
  return ( a2*c_avg + b2*n_avg + c2*m_avg + 0.117 )*( sqrt(n) - sqrt(n0) );

}



// convert CH4 concentration in ppb to radiative forcing...:
func ch4_c2f( m, n, m0=, n0= )
{
  
  // default values:
  if( is_void(n) )
    n = 270.;            // preindustrial N2O concentration (ppb)
  if( is_void(m0) )
    m0 = 700.;           // preindustrial CH4 concentration (ppb)
  if( is_void(n0) )
    n0 = 270.;           // preindustrial N2O concentration (ppb)

  if( n<200. | n>525. | m<340. | m>3500. )
    exit, "[ch4_c2f] Greenhouse-gas concentrations outside valid range!";

  
  // constants:
  a3 = -1.3e-6;          // (W/m2/ppb)
  b3 = -8.2e-6;          // (W/m2/ppb)

  // calculate and return radiative forcing:
  m_avg = 0.5*(m+m0);
  n_avg = 0.5*(n+n0);
  return ( a3*m_avg + b3*n_avg + 0.043 )*( sqrt(m) - sqrt(m0) );
  
}





//  The following were copied from maunder_forcings.i and renamed to
//  ..._old for comparison and double-checks.

// convert CO2 concentration in ppm to radiative forcing...:
func co2_c2f_old( c, c0= )
{
  a = 5.35;  // conversion factor
  if( is_void(c0) )
    c0 = 278.; // preindustrial CO2 concentration (ppm)
  return a*log(c/c0);
}

// ... and backwards:
func co2_f2c_old( f, a=, c0= )
{
  if( is_void(a) )
    a = 5.35;  // conversion factor
  if( is_void(c0) )
    c0 = 278.; // preindustrial CO2 concentration (ppm)
  return c0*exp(f/a);
}

// convert CH4 concentration in ppb to radiative forcing...:
func ch4_c2f_old( c, b=, m0=, n0= )
{
  if( is_void(b) )
    b = 0.036; // conversion factor
  if( is_void(m0) )
    m0 = 700.; // preindustrial CH4 concentration (ppb)
  if( is_void(n0) )
    n0 = 270.; // preindustrial N2O concentration (ppb)
  return b*(sqrt(c)-sqrt(m0)) - ( _f_(c,n0) - _f_(m0,n0) );
}

// convert N2O concentration in ppb to radiative forcing...:
func n2o_c2f_old( c, e=, m0=, n0= )
{
  if( is_void(e) )
    e = 0.12;  // conversion factor
  if( is_void(m0) )
    m0 = 700.; // preindustrial CH4 concentration (ppb)
  if( is_void(n0) )
    n0 = 270.; // preindustrial N2O concentration (ppb)
  return e*(sqrt(c)-sqrt(n0)) - ( _f_(m0,c) - _f_(m0,n0) );
}

func _f_( m, n )
{
  return 0.47*log( 1. + 2.01e-5*(m*n)^0.75 + 5.31e-15*m*(m*n)^1.52 );
}
