require, "mathlib.i";
require, "matrix.i";


func ssa ( t, x, M, &E, &V, &A, &R, &C, method= ) 
/* DOCUMENT ssa, t, x, M, &E, &V, &A, &R, &C, method=

   SSA - driver routine to perform Singular Spectrum Analysis

   INPUT:

   t, x    Time series
   M       Embedding dimension.
   method= (optional) method of calculating the covariance matrix:
           'unbiased' (N-k weighted) (default)
           'biased'   (N-weighted or Yule-Walker)  
           'BK'       (Broomhead/King type estimate)

   OUTPUT: 

   E       Eigenfunction (T-EOF) matrix in standard form
   V       Vector containing variances (unnormalized eigenvalues)
   A       Matrix of principal components
   R       Matrix of reconstructed components
   C       Covariance matrix

   See Vaut ard, Yiou, and Ghil, Physica D 58, 95-126, 1992.

   HISTORY:  
  
   Written by Eric Breitenberger.     Version date 1/22/96
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)

*/
{

  if(is_void(method))
    method = "unbiased";

  // call ssaeig to perform SSA
  ssaeig, t, x, M, E, V, C, method=method;

  // compute principal components
  A = pc( x, E );

  // compute reconstructed components
  R = rc( A, E );

}



func ssaeig ( t, x, M, &E, &V, &C, method= )
/* DOCUMENT ssaeig, t, x, M, &E, &V, &C, method=

   SSAEIG - starts an SSA of series 't, x', for embedding dimension 'M'.

   INPUT:

   t, x    Time series
   M       Embedding dimension.
   method= (optional) method of calculating the covariance matrix:
           'unbiased' (N-k weighted) (default)
           'biased'   (N-weighted or Yule-Walker)  
           'BK'       (Broomhead/King type estimate)

   OUTPUT: 

   E       Eigenfunction (T-EOF) matrix in standard form
   V       Vector containing variances (unnormalized eigenvalues)
   C       Covariance matrix

   E and V are ordered from large to small.
   See section 2 of Vautard, Yiou, and Ghil, Physica D 58, 95-126, 1992.

   HISTORY:  

   Written by Eric Breitenberger.    Version date 1/22/96
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)

*/
{

  if( is_void(method) )
    method = "unbiased";
  
  N = numberof(x);
  if( M>=N-M+1 )
    error("Hey! Too big a lag!");
  
  if(method!="BK"){
    c = ac ( x, M-1, method=method ); // calculate autocovariance estimates
    C = toeplitz(c);                  // create Toeplitz matrix (trajectory matrix)
  }
  else 
    C = bk( x, M );                   // Broomhead/King estimate
  
  // calculate eigenvectors, values of C:
  E = [];
  L = SVdec( C, E );

  // create sorted eigenvalue vector
  sortind = sort(-L);
  V = L(sortind);

  V = -transpose(V);
  E=E(,sortind);                      // sort eigenvector matrix
  
}



func ssatrend( t, x, M, &R, &V, &SE_R, &trendidx, &tfilt, 
               sE_x=, ssamethod=, extrapmethod= )

/* DOCUMENT ssatrend, t, x, M, &R, &V, &SE_R, &trendidx, &tfilt, 
   sE_x=, ssamethod=, extrapmethod=
   
   Finds the SSA non-linear trend from a time series.

   INPUT:

   t, x          Time series
   M             Embedding dimension
   SE_x=         (optional) standard error of a measurement in x
   ssamethod=    (optional) method of calculating the covariance matrix:
                 'unbiased' (N-k weighted) (default)
                 'biased'   (N-weighted or Yule-Walker)  
                 'BK'       (Broomhead/King type estimate)
   extrapmethod= (optional) "minimumroughness" (default) or "minimumslope"


   OUTPUT:

   R             the nonlinear trend (a specific reconstructed component where 
                 x has succesively padded with the linear trend of the last M points.
                 For details check source). Note that the mean(x) has been added to 
                 the RC (so that it is easy to plot on the same plot as the original
                 timeseries)
   V             the variance explained by R (as a fraction of the total variance)
   SE_R          standard error of the R (how much of sE_x is remaining in R). 
   trendidx      the index of the rc corresponding to the trend
   tfilt         symmetric reconstructive SSA filter. (what is convoluted on x)
 
   HISTORY:

   (c) Aslak Grinsted 2004 

   Converted from Matlab to Yorick by Georg Feulner (2009)

*/

{
  
  if(!ssamethod)
    ssamethod="unbiased";

  if(!extrapmethod)
    extrapmethod="minimumroughness";

  meanx = avg(x);
  
  if(is_void(sE_x))
    sE_X = [];

  if(numberof(sE_x)==1)
    sE_x = array(sE_x, numberof(x) );

  if(numberof(sE_x)<1){
    sE_x = array( 0., numberof(x) );
    noerrorbar = 1;
  }

  if( avg(sE_x)>stddev(x) )
    write, "Warning! sE_x > stddev(x)!!!";

  n = numberof(x);


  if(numberof(M)==1){

    E = V = A = R = C = [];
    ssa, t, x, M, E, V, A, R, C, method=ssamethod;


    // find trend idx: trend would be non oscillatory and therefore have 
    // a large sum in the index 

    EE = abs(sum(E))/sum(abs(E));
    mx = max(EE);
    trendidx = where( EE==mx );


    // throw away everything but the trend:
    E = E(,trendidx);
    V = V(trendidx)/sum(V); // convert to percent

    if(V<0.1)
      write, "Warning! SSA trend accounts for less than 10% of the variance!";

    R = R(,trendidx) + meanx;

    tfilt = convol(E,E(indgen(numberof(E):1:-1)))/M;

  }
  else{
    tfilt = M;
    V =0.;
    trendidx = 0.;
    M = ceil(length(tfilt)/2); // used for padding.
  }



  tfilt = tfilt/sum(tfilt); // NORMALIZE!!!!!!!



  // GF: commented in Stefan's version
 
  // succesively pad the series with values from the trend and reconstruct
  // again to get better end estimates.
  paddedX = x;
  for( ii=1; ii<=M; ii++ ){
    leftR  = avg(paddedX(1:M));
    rightR = avg(paddedX(-(M-1):0));
    xxx = [leftR];
    grow, xxx, paddedX, rightR;
    paddedX = xxx;
  }
  
  mp = M;

  if( extrapmethod="minimumroughness" ){

    idx = indgen(mp);
    pleft = fitpoly( 1, idx, x(idx) );

    epleft = sqrt(max(sE_x(idx))^2+stddev(detrend(x(idx)))^2);

    idx += n-mp;
    pright = fitpoly( 1, idx, x(idx) );
    epright = sqrt(max(sE_x(idx))^2+stddev(detrend(x(idx)))^2);

    paddedX = pleft(1) + pleft(2)*indgen(-(M-1):0:1);
    grow, paddedX, x;
    grow, paddedX, pright(1) + pright(2)*indgen(n+1:n+M:1);
    
  }
  else{

     idx = indgen(mp);
     pleft = avg(x(idx));
     epleft=sqrt(max(sE_x(idx))^2+stddev(x(idx))^2);
     idx = indgen(n-mp+1:n:1);
     pright = avg(x(idx));
     epright=sqrt(max(sE_x(idx))^2+stddev(x(idx))^2);
     paddedX = array( pleft, mp );
     grow, paddedX, x, array( pright, mp );

  }


  R = convol( paddedX, tfilt, n0=2*M, n1=2*M+n-1 );

  
  // What happens to whitenoise if it goes through this tfilt?
  if(noerrorbar)
    sE_R = [];
  else{
    // pad sE_x
    sE_x = array( epleft, M );
    grow, sE_x, sE_x, array( epright, M );


    // GF: I am not sure about this if statement, which is always true, of course...
    if(1){
      // MONTE CARLO WAY
      mcreal = array( 0., numberof(x), 1000 );
      for( ii=1; ii<=dimsof(mcreal)(3); i++ ){
        temp = convol( paddedX+random(numberof(paddedX))*sE_x, tfilt,
                       n0=2*M, n1=2*M+n-1 );
        nnn = floor( (numberof(temp)-n)/2. );
        mcreal(,ii) = temp(nn+1:nn+n);
      }
      sE_R = stddev(mcreal);
    }
    else{
      et = random(100000); // 100000 might be overkill but conv is so fast anyway.
      et = convol( et, tfilt, n0=2*M, n1=2*M+numberof(et)-1 );
      sE_R = stddev(et);

      sE_R = convol( sE_x, tfilt/sum(tfilt), n0=2*M, n1=2*M+n-1 )*sE_R;

      nnn = floor( (numberof(sE_R)-n)/2. );
      sE_R = sE_R(nn+1:nn+n);

    }

  }

}


func  ac ( x, k, method= )
/* DOCUMENT c = ac( x, k, method= )

   Calculates the auto-covariance for series x out to lag k. The 
   result is output in c, which has k+1 elements. The first element
   is the covariance at lag zero; succeeding elements 2:k+1 are the
   covariances at lags 1 to k.
 
   INPUT:

   x       Time series
   k       Lag
   method= "biased"   (N-weighted or Yule-Walker)  
           "unbiased" (N-k weighted - this is the default)
           "BK"       (Broomhead/King type estimate)
 
           Note that the BK method is of limited use unless you are computing
           the full covariance matrix - see BK.M.                    
 
   OUTPUT:
   
   c       Auto-covariance

   HISTORY:

   Written by Eric Breitenberger.   Version date 1/11/96
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)

*/
{
  N = numberof(x);
  if(k>=N)
    error, "Too big a lag!";

  x = x-avg(x); // center the series.
  c = array( 0., k+1 );

  if(is_void(method))
    method="unbiased";

  if(method="unbiased")
    for( i=1; i<=k+1; i++ )
      c(i) = sum(x(1:N-i+1)*x(i:N))/(N-i+1);
  
  else if(method="biased"){
    for( i=1; i<=k+1; i++ )
      c(i) = sum(x(1:N-i+1)*x(i:N));
    c = c/N;
  }

  else if( method="BK"){
    for( i=1; i<=k+1; i++ )
      c(i) = sum(x(1:N-k)*x(i:N-k+i-1));
    c = c/(N-k);
  }
  
  else
    error, "Improper specification of method.";

  return c;

}


func bk( X, M )
/* DOCUMENT  C = bk( X, M )
   
   Computes a Broomhead/King type covariance matrix for the
   centered multichannel process X, for embedding dimension 'M'.
   Each of the columns of X is a time series.
 
   INPUT:

   X     Time series. Note that X must be centered.
 
   OUTPUT:

   C     Lag-covariance matrix.
 
   HISTORY:

   Written by Eric Breitenberger.    Version date 1/11/96
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)
   
*/
{
  d = dimsof( X );
  N = d(2);
  L = d(3);
  
  if( M*L>=N-M+1)
    write, "Warning: Covariance matrix may be ill-conditioned.'";

  // Create the trajectory matrix: this code differs from 
  // that in EEOF only in the column ordering.
  T = array( 0., N-M+1, M*L );
  index = span( 0, (L-1)*M, L );

  for( i=1; i<=M; i++ )
    T(,index+i) = X(i:N-M+i,);
  
  C = (transpose(T)*T)/(N-M+1);

  return C;

}


func pc( x, E )
/* DOCUMENT p = PC( x, E )

   Calculates the principal components of the series x
   from the eigenfunction matrix E (output from ssaeig).

   INPUT:

   x      Time series
   E      Eigenfunction matrix (output from ssaeig)

   OUTPUT:

   A      Principal components matrix (N-M+1 x M)
 
   See section 2.4 of Vautard, Yiou, and Ghil, Physica D 58, 95-126, 1992.
 
   HISTORY:

   Written by Eric Breitenberger.    Version date 5/20/95
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)
   
*/
{ 
  d = dimsof(x);
  if( d(1)>1 )
    error, "x must be a vector.";
  N = d(2);

  x =x-avg(x);
  d = dimsof(E);
  
  if( d(1)!=2 )
    error, "E is improperly dimensioned";
  M = d(2);
  c = d(3);
  
  if( M!=c)
    error, "E is improperly dimensioned";

  A = array( 0., N-M+1, M );

  for( i=1; i<=N-M+1; i++ ){                 
    w = x(i:i+M-1);          
    A(i,)=w(+)*E(+,);
  }

  return A;

}


func rc ( A, E )
/* DOCUMENT R = rc( A, E )

   This function calculates the 'reconstructed components' using the 
   eigenvectors (E, from ssaeig) and principal components (A, from pc).
   R is N x M, where M is the embedding dimension used in ssaeig.

   INPUT:

   A     Principal components (from pc)
   E     Eigenvectors (from ssaeig)

   OUTPUT:

   R     Reconstructed components (N x M, where M embedding dimensions)

   See section 2.5 of Vautard, Yiou, and Ghil, Physica D 58, 95-126, 1992.
 
   HISTORY:

   Written by Eric Breitenberger.   Version date 5/18/95
   Please send comments and suggestions to eric@gi.alaska.edu   

   Converted from Matlab to Yorick by Georg Feulner (2009)
   
*/
{
  de = dimsof(E);
  if( de(1)!=2 )
    error, "E is improperly dimensioned";
  M = de(2);
  c = de(3);
  if( M!=c )
    error, "E is improperly dimensioned";

  da = dimsof(A);
  if( da(1)!=2 )
    error, "A is improperly dimensioned";
  ra = da(2);
  ca = da(3);
  if( ca!=M )
    error, "A is improperly dimensioned";
  
  N = ra+M-1;  // Assumes A has N-M+1 rows.


  R = array( 0., N, M );
  Z = array( 0., M-1, M );
  // original: A=[A' Z'];
  aa = transpose(A);
  grow, aa, transpose(Z);
  A = transpose(aa);

  // Calculate RCs
  ncon = int( floor(M/2) );
  for( k=1; k<=M; k++ )
    R(,k) = convol( A(,k), E(,k), n0=ncon+1, n1=N+ncon )/M;

  // Adjust first M-1 rows and last M-1 rows
  for( i=1; i<M; i++ ){
    R(i,)=R(i,)*(M/i);
    R(N-i+1,)=R(N-i+1,)*(M/i);
  }


  return R;

}


func toeplitz( v )
/* DOCUMENT t = toeplitz( v )
   
   Creates the Toeplitz matrix from vector v

   Georg Feulner (2009)
*/
{
  n = numberof(v);
  T = array( 0., n, n );
  for( i=1; i<=n; i++ ){
    T(i:n,i) = v(1:n-i+1);
    T(i,i:n) = v(1:n-i+1);
  }
  return T;
}
